/* MouseWheelHandling.m created by lindberg on Tue 07-Dec-1999 */
/*
 * Copyright (c) 2002 Carl Lindberg and Quinten Mercer
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#import "MouseWheelHandling.h"
#import "RCWinMouseWheel.h"
#import <AppKit/NSClipView.h>

@implementation NSScrollView (RCMouseWheel)

- (void)__scrollVertBy:(float)delta
{
    NSView *docView = [self documentView];
    NSRect visibleRect = [docView visibleRect];
    NSRect docFrame = [docView frame];

    visibleRect.origin.y -= delta;
    if (NSMaxY(visibleRect) > NSMaxY(docFrame))
        visibleRect.origin.y = docFrame.size.height - visibleRect.size.height;
    if (NSMinY(visibleRect) < NSMinY(docFrame))
        visibleRect.origin.y = 0.0;

    [docView scrollRectToVisible:visibleRect];
}

- (void)__scrollHorizBy:(float)delta
{
    NSView *docView = [self documentView];
    NSRect visibleRect = [docView visibleRect];
    NSRect docFrame = [docView frame];

    visibleRect.origin.x += delta;
    if (NSMaxX(visibleRect) > NSMaxX(docFrame))
        visibleRect.origin.x = docFrame.size.width - visibleRect.size.width;
    if (NSMinX(visibleRect) < NSMinX(docFrame))
        visibleRect.origin.x = 0.0;

    [docView scrollRectToVisible:visibleRect];
}

/*"
 * The handler for NSScrollView. Currently just does simplistic scrolling of the
 * docView.
"*/
- (BOOL)mouseWheel:(NSEvent *)event
{
    int numPixels = [event mouseWheelDelta];

    /*
     * If the 'Command' key is pressed while scrolling the scrolling slows
     * down to 1/10 the original speed.  Creating a fine granularity scroll.
     * If both the 'Command' and 'Shift' keys are pressed while scrolling the
     * scrolling speeds up to 10 times the original speed, thus allowing the
     * user to jump through pages of data quickly.
     */
    if ([event modifierFlags] & NSCommandKeyMask)
    {
        if ([event modifierFlags] & NSShiftKeyMask)
            numPixels = (numPixels * 10);
        else
            numPixels = (numPixels / 10);
    }

    if ([self hasHorizontalScroller] && ![self hasVerticalScroller]) {
        [self __scrollHorizBy:(-numPixels / 2)];
    }
    else {
        if (![[self contentView] isFlipped])
            numPixels = -numPixels;
        [self __scrollVertBy: (numPixels / 2)];
    }
    return YES;
}

- (BOOL)verticalScroll:(NSEvent *)event
{
    float scrollBy = -[event scrollIncrement];

    /*
     * For the ScrollLines scrollType we take into consideration any
     * modifier keys that are depressed.  (actually just control and
     * shift for now)
     */
    switch ([event scrollType])
    {
        case ScrollLines:
            scrollBy *= 30;  //30 pixels per line.  Maybe we should be using [self lineHeight] here.
            if ([event modifierFlags] & NSCommandKeyMask)
            {
                if ([event modifierFlags] & NSShiftKeyMask)
                    scrollBy *= 10;
                else
                    scrollBy /= 10;
            }
            break;
        case ScrollPages: scrollBy *= [[self documentView] visibleRect].size.height; break;
        case ScrollMax:   scrollBy *= 30000; break;  //cheesy
        default: return NO;
    }

    if (![[self contentView] isFlipped])
        scrollBy = -scrollBy;
    [self __scrollVertBy:scrollBy];

    return YES;
}

- (BOOL)horizontalScroll:(NSEvent *)event
{
    float scrollBy = [event scrollIncrement];

    /*
     * For the ScrollLines scrollType we take into consideration any
     * modifier keys that are depressed.  (actually just control and
     * shift for now)
     */
    switch ([event scrollType])
    {
        case ScrollLines:
            scrollBy *= 30;
            if ([event modifierFlags] & NSCommandKeyMask)
            {
                if ([event modifierFlags] & NSShiftKeyMask)
                    scrollBy *= 10;
                else
                    scrollBy /= 10;
            }
            break;
        case ScrollPages: scrollBy *= [[self documentView] visibleRect].size.width; break;
        case ScrollMax:   scrollBy *= 30000; break;  //cheesy
        default: return NO;
    }

    [self __scrollHorizBy:scrollBy];
    return YES;
}

@end

static BOOL WrapCombos = NO;
static inline void _setupComboDefaults()
{
    static BOOL initted = NO;
    if (!initted) {
        initted = YES;
        WrapCombos = [[NSUserDefaults standardUserDefaults] boolForKey:@"RCMouseWheelComboBoxWraps"];
    }
}

@implementation NSComboBox (RCMouseWheel)

/*"
 * The handler for NSComboBox. Iterates through the items.
"*/
- (BOOL)mouseWheel:(NSEvent *)event
{
    int currIndex = [self indexOfSelectedItem];
    int numItems  = [self numberOfItems];
    int itemDelta;
    int newIndex;

    if (currIndex < 0) // index is sometimes -1 if user never set it with mouse
    {
        NS_DURING
            if (![self usesDataSource])
                currIndex = [self indexOfItemWithObjectValue:[self objectValue]];
            else if ([self dataSource] != nil)
                currIndex = [[self dataSource] comboBox:self indexOfItemWithStringValue:[self stringValue]];
        NS_HANDLER
            currIndex = -1;
        NS_ENDHANDLER
    }

    if (currIndex < 0 || numItems < 2) return YES;

    _setupComboDefaults();

    itemDelta = ([event mouseWheelDelta] / MouseWheelNotchLevel) * -1;
    newIndex = currIndex + itemDelta;

    /* We can wrap off the top to the bottom, or just stay at the top */
    if (WrapCombos)
    {
        if (newIndex >= numItems) newIndex = numItems - newIndex;
        while (newIndex < 0) newIndex += numItems;
    }
    else
    {
        if (newIndex < 0) newIndex = 0;
        if (newIndex >= numItems) newIndex = numItems - 1;
    }

    /* Bail if we don't need to do anything */
    if (newIndex == currIndex) return YES;

    // Change selection, while sending appropriate notifications/actions.
    // NSComboBoxSelectionDidChangeNotification is posted by -selectItemAtIndex:.
    [[NSNotificationCenter defaultCenter]
        postNotificationName:NSComboBoxSelectionIsChangingNotification object:self userInfo:nil];
    [self selectItemAtIndex:newIndex];
    if ([self action] != NULL)
        [self sendAction:[self action] to:[self target]];

    return YES;
}

@end

