/* WinNTMouseWheel.m created by mercer on Mon 06-Dec-1999 */
/*
 * Copyright (c) 2002 Carl Lindberg and Quinten Mercer
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#import "RCWinMouseWheel.h"
#import <Foundation/NSUserDefaults.h>
#import <AppKit/NSView.h>
#import <AppKit/NSWindow.h>
#import <AppKit/NSApplication.h>
#import <AppKit/NSDPSServerContext.h>
#import <windows.h>

static void InstallMouseWheelHandling();
static BOOL RCMouseWheelEnabled = YES;
static BOOL StartWithFirstResponder = NO;
static BOOL TraceAllWindowsEvents = NO;
static BOOL TraceWindowsEvents = NO;
static BOOL CoalesceEvents = YES;

typedef BOOL (*SendEventIMP)(id, SEL, NSEvent*);
static SendEventIMP windowSendEvent;
static IMP RCOverrideInstanceMethodWithSelector(Class aClass, SEL oldSelector, SEL newSelector);

static void ReplaceWindowProc(LPCSTR className, WNDPROC newProc, WNDPROC *oldProcPointer);
static WNDPROC OriginalDefaultEventHandler;
static WNDPROC OriginalDialogEventHandler;
extern LRESULT APIENTRY RCDefaultEventHandler(HWND window, UINT msg, WPARAM wparam, LPARAM lparam);
extern LRESULT APIENTRY RCDialogEventHandler(HWND window, UINT msg, WPARAM wparam, LPARAM lparam);


@implementation RCWinMouseWheel
/* 
 * Install the mouse event stuff upon the InputManager being loaded. Do it in
 * -init as well just in case +load fails.
 */
+ (void)load
{
    InstallMouseWheelHandling();
}
- init
{
    InstallMouseWheelHandling();
    return self;
}
@end

@implementation NSResponder (RCMouseWheel)

- (BOOL)mouseWheel:(NSEvent *)event       { return NO; }
- (BOOL)verticalScroll:(NSEvent *)event   { return NO; }
- (BOOL)horizontalScroll:(NSEvent *)event { return NO; }

@end

@interface NSEvent (RCMouseWheelPrivate)
// for use by coalescing code
- (void)setMouseWheelDelta:(int)value;
- (void)setScrollIncrement:(int)value;
@end

@implementation NSEvent (RCMouseWheel)

/*"
 * Returns delta for this mouse wheel event; the distance the mouse wheel
 * traveled. It is most likely an even multiple of MouseWheelNotchLevel, which
 * each level representing a turn of one "notch" on the mouse wheel. A positive
 * value means the wheel was rotated forward (away from the user, or up), and a
 * negative value means it was rotated backwards (towards the user, or down).
 * The return value is only valid for a mouse wheel event (type NSAppKitDefined,
 * subtype MouseWheelSubtype).  The value is actually stored in -#data1.
"*/
- (int)mouseWheelDelta
{
    return [self data1];
}

- (ScrollType)scrollType
{
    return [self data1];
}

- (int)scrollIncrement
{
    return [self data2];
}

- (void)setMouseWheelDelta:(int)value
{
    _data.misc.data1 = value;
}
- (void)setScrollIncrement:(int)value
{
    _data.misc.data2 = value;
}

@end


static void InstallMouseWheelHandling()
{
    static BOOL installed = NO;
    if (!installed)
    {
        NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
        NSDictionary *registration = [NSDictionary dictionaryWithObjectsAndKeys:
            @"YES", @"RCMouseWheelEnabled",
            @"YES", @"RCMouseWheelCoalesceEvents",
            nil];

        installed = YES;

        /* Set up defaults */
        // we might also want a setting that starts with first responder, but only
        // operates if the mouse is within that view's bounds.  That seems to be
        // more like IE does it.  Course other apps work differently.
        [defaults registerDefaults:registration];
        RCMouseWheelEnabled     = [defaults boolForKey:@"RCMouseWheelEnabled"];
        StartWithFirstResponder = [defaults boolForKey:@"RCMouseWheelUseFirstResponder"];
        TraceAllWindowsEvents   = [defaults boolForKey:@"RCMouseWheelTraceAllEvents"];
        TraceWindowsEvents      = [defaults boolForKey:@"RCMouseWheelTraceEvents"];
        CoalesceEvents          = [defaults boolForKey:@"RCMouseWheelCoalesceEvents"];

        if (!RCMouseWheelEnabled) return;

        /* Replace NSWindow's -sendEvent: with our version */
        windowSendEvent = (SendEventIMP)RCOverrideInstanceMethodWithSelector([NSWindow class],
                               @selector(sendEvent:), @selector(mouseWheelSendEvent:));

        /* 
         * Replace the WNDPROC in AppKit's main window classes with ours. The
         * window class names are courtesy of strings.exe... there is also a
         * "SplashWindowClassName" in there, but we don't care about modifying
         * splash panels....
         */
        ReplaceWindowProc("NSDefaultWindowClass",
                          RCDefaultEventHandler,
                          &OriginalDefaultEventHandler);
        ReplaceWindowProc("NSDialogWindowClass",
                          RCDialogEventHandler,
                          &OriginalDialogEventHandler);
    }
}

/* Replaces a Windows WNDPROC setting for a window class */
static void ReplaceWindowProc(LPCSTR className, WNDPROC newProc, WNDPROC *oldProcPointer)
{
    HINSTANCE appInstance = [NSApp applicationHandle];
    WNDCLASS windowClass;

    /* Get the current settings */
    GetClassInfo(appInstance, className, &windowClass);
    *oldProcPointer = windowClass.lpfnWndProc;
    windowClass.lpfnWndProc = newProc;

    /* Try to unregister the old class and register the new one. This will fail
       if there are any open windows. */
    if (UnregisterClass(className, appInstance) != 0)
    {
        RegisterClass(&windowClass);
    }
    else
    {
        /* 
         * There are open windows, so we have to use the SetWindowLong()
         * function to reset the window proc on each window of the current
         * class. We also call SetClassLong() to update the window class so
         * future windows will have our new function too. It would be better to
         * do it this way in all cases instead of unregistering, but we need an
         * active window to be able to call SetClassLong()!
         */
        NSArray *allWindows = [NSApp windows];
        int i, count = [allWindows count];
        BOOL doneClass = NO;

        for (i=0; i<count; i++)
        {
            NSWindow *window = [allWindows objectAtIndex:i];
            HWND     handle = [window windowHandle];
            WNDPROC  currentProc = (WNDPROC)GetWindowLong(handle, GWL_WNDPROC);

            if (currentProc == *oldProcPointer)
            {
                SetWindowLong(handle, GWL_WNDPROC, (long)newProc);
                if (!doneClass) {
                    SetClassLong(handle, GCL_WNDPROC, (long)newProc);
                    doneClass = YES;
                }
            }
        }
    }
}


/**** Handle the events as they come through from Windows ****/

static NSDPSContext *CurrentDPSContext()
{
    static NSDPSServerContext *serverContext = nil;
    if (serverContext == nil) {
        id aContext = [NSDPSServerContext currentContext];
        if ([aContext isKindOfClass:[NSDPSServerContext class]])
            serverContext = aContext;
    }
    return serverContext;
}

static BOOL CanCoalesceMouse;
static int  CoalesceMouseValue;
static BOOL CanCoalesce[3];
static int  CoalesceValue[3];
static NSTimeInterval CoalesceTime = -100000000.0;

static LRESULT CommonEventHandler(HWND window, UINT msg, WPARAM wparam, LPARAM lparam,
                                  WNDPROC origHandler)
{
    /*
     * Only the low word is the message; the high word is reserved for the system.
     * The high word may already be stripped at this point, but we'll be careful.
     */
    msg = LOWORD(msg);

    if (TraceAllWindowsEvents)
    {
        printf("event type: %#x, wparam: %#x (%d, %d), lparam: %#x (%d, %d), extra: %#x\n", msg,
               (unsigned)wparam, (short)HIWORD(wparam), (short)LOWORD(wparam),
               (unsigned)lparam, (short)HIWORD(lparam), (short)LOWORD(lparam),
               (int)GetMessageExtraInfo());
    }

/* 
 * The WinNT #defines for their mouse wheel stuff. These should be defined in
 * the normal headers, but it seems as though they have been conditionally
 * #ifdef'ed out, so we need to define them ourselves.
 */
#ifndef WHEEL_DELTA
#define WHEEL_DELTA 120
#endif
#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL 0x020A
#endif
// For Logitech WheelMouse, according to other MouseWheel IM package.
// I'm not sure we need this, since we get events after they've been
// processed by the mouse driver, but I don't think it can hurt.
// I do see 0xC0C1 and 0xC0CB events with Logitech stuff, and I don't
// know what they are...
// Hm.  Upon more research, messages in the range 0xC000 and 0xFFFF
// are for when apps use RegisterWindowMessage().  Values are guaranteed
// to be unique throughout the system, so it's unlikely that values in
// this range mean the same thing from run to run, or at least boot to
// boot.  So... I'm guessing this value is not really a constant.
// If we knew the name Logitech uses for the event we may be able to
// figure it out...
#ifndef LM_MOUSEWHEEL
#define LM_MOUSEWHEEL 0xC10B
#endif

    if (msg == WM_MOUSEWHEEL /*|| msg == LM_MOUSEWHEEL*/)
    {
        NSWindow *windowObject = [NSApp windowWithWindowHandle:window];
        NSScreen *screen = [windowObject screen];
        NSPoint screenPoint, windowPoint;
        short modifierFlags;
        int   eventFlags = 0;
        short rawWheelDelta;
        int   adjustedDelta;
        NSEvent *event;
        NSTimeInterval eventTime = [NSDate timeIntervalSinceReferenceDate];

        if (screen == nil) screen = [NSScreen mainScreen];
        screenPoint.x = LOWORD(lparam);
        screenPoint.y = [screen frame].size.height - HIWORD(lparam);
        modifierFlags = LOWORD(wparam);
        rawWheelDelta = HIWORD(wparam);

        /* Convert to our notch level from Windows' WHEEL_DELTA */
        adjustedDelta = (rawWheelDelta * MouseWheelNotchLevel) / WHEEL_DELTA;

        /* 
         * Attempt coalescing. Ignore events > 1 second apart, as that's likely
         * because of an event getting posted but never delivered, or at least
         * NSWindow's sendEvent not getting called for it. Happens to events
         * sent to popup windows (NSPopUpButton/NSComboBox).
         */
        if (CoalesceEvents && CanCoalesceMouse && (eventTime - CoalesceTime) < 1.0 ) {
            if (TraceWindowsEvents)
                printf("Coalescing mouse event (%d)\n", adjustedDelta);
            CoalesceMouseValue += adjustedDelta;
            return 0;
        }

        /* Convert the mouse location to be relative from the window */
        windowPoint = [windowObject convertScreenToBase:screenPoint];

        /*
         * Set event flags.  NOTE: very simplistic.  Does not set many
         * other mask bits (caps lock, num lock, etc.) and does not take
         * into account the NSModifierFlagMapping default.  We also might
         * want to add specialized flags for the various mouse buttons
         * being down, which is also in the modifierFlags value.
         */
        if (modifierFlags & MK_CONTROL)
            eventFlags |= NSCommandKeyMask;
        if (modifierFlags & MK_SHIFT)
            eventFlags |= NSShiftKeyMask;
        if (GetKeyState(VK_MENU) & 0x80)
            eventFlags |= NSAlternateKeyMask;

        event = [NSEvent otherEventWithType:NSApplicationDefined
                                   location:windowPoint
                              modifierFlags:eventFlags
                                  timestamp:0
                               windowNumber:[windowObject windowNumber]
                                    context:CurrentDPSContext()
                                    subtype:MouseWheelSubtype
                                      data1:adjustedDelta
                                      data2:0];

        CanCoalesceMouse = YES;
        CoalesceMouseValue = adjustedDelta;
        CoalesceTime = eventTime;

        if (TraceWindowsEvents)
            printf("Posting mouse wheel %#x\n", (unsigned)event);

        [NSApp postEvent:event atStart:NO];
        return 0;
    }
    else if (msg == WM_VSCROLL || msg == WM_HSCROLL)
    {
        NSWindow   *windowObject = [NSApp windowWithWindowHandle:window];
        NSScreen   *screen = [windowObject screen];
        POINT      winScreenPoint;
        NSPoint    screenPoint, windowPoint;
        ScrollType scrollType;
        NSEvent    *event;
        int        eventSubtype = (msg == WM_VSCROLL)? VerticalScrollSubtype : HorizontalScrollSubtype;
        int        adder = 0;
        NSTimeInterval eventTime = [NSDate timeIntervalSinceReferenceDate];

        if (screen == nil) screen = [NSScreen mainScreen];
        GetCursorPos(&winScreenPoint);
        screenPoint.x = winScreenPoint.x;
        screenPoint.y = [screen frame].size.height - winScreenPoint.y;
        /* Convert the mouse location to be relative from the window */
        windowPoint = [windowObject convertScreenToBase:screenPoint];

        switch (LOWORD(wparam))
        {
            case SB_LINEDOWN: scrollType = ScrollLines; adder =  1; break; // also SB_LINERIGHT
            case SB_LINEUP:   scrollType = ScrollLines; adder = -1; break; // also SB_LINELEFT
            case SB_PAGEDOWN: scrollType = ScrollPages; adder =  1; break; // also SB_PAGERIGHT
            case SB_PAGEUP:   scrollType = ScrollPages; adder = -1; break; // also SB_PAGELEFT
            case SB_BOTTOM:   scrollType = ScrollMax;   adder =  1; break; // also SB_RIGHT
            case SB_TOP:      scrollType = ScrollMax;   adder = -1; break; // also SB_LEFT
            default:
                return 0;  // we don't care about other values
        }

        if (CoalesceEvents && CanCoalesce[scrollType] && (eventTime - CoalesceTime) < 1.0)
        {
            if (scrollType == ScrollMax)
            {
                if (adder == CoalesceValue[scrollType]) {
                    if (TraceWindowsEvents) printf("Coalescing max scroll event (%d)\n", adder);
                    //we already have a max event pending; just discard this one.
                    return 0;
                }
            }
            else
            {
                if (TraceWindowsEvents) printf("Coalescing %d scroll event (%d)\n", scrollType, adder);
                CoalesceValue[scrollType] += adder;
                return 0;
            }
        }

        CoalesceValue[scrollType] = adder;
        CanCoalesce[scrollType] = YES;
        CoalesceTime = eventTime;

        event = [NSEvent otherEventWithType:NSApplicationDefined
                                   location:windowPoint
                              modifierFlags:0
                                  timestamp:0
                               windowNumber:[windowObject windowNumber]
                                    context:CurrentDPSContext()
                                    subtype:eventSubtype
                                      data1:scrollType
                                      data2:adder];

        if (TraceWindowsEvents)
            printf("Posting %d scroll\n", eventSubtype);
        [NSApp postEvent:event atStart:NO];
        return 0;
    }
    else
    {
        //return origHandler(window, msg, wparam, lparam);
        //CallWindowProc() is the more official Win32 way it seems.
        #ifdef STRICT
        return CallWindowProc(origHandler, window, msg, wparam, lparam);
        #else
        return CallWindowProc((FARPROC)origHandler, window, msg, wparam, lparam);
        #endif
    }
}

/* The real handlers */
extern LRESULT RCDefaultEventHandler(HWND window, UINT msg, WPARAM wparam, LPARAM lparam)
{
    return CommonEventHandler(window, msg, wparam, lparam, OriginalDefaultEventHandler);
}

extern LRESULT RCDialogEventHandler(HWND window, UINT msg, WPARAM wparam, LPARAM lparam)
{
    return CommonEventHandler(window, msg, wparam, lparam, OriginalDialogEventHandler);
}


/* Handle the event when sent to a window.  Find someone that implements -mouseWheel. */
@implementation NSWindow (RCMouseWheel)

- (BOOL)mouseWheelSendEvent:(NSEvent *)theEvent
{
    if ([theEvent type] == NSApplicationDefined && [theEvent subtype] == MouseWheelSubtype)
    {
        id responder;

        if (TraceWindowsEvents)
            printf("Handling mouse wheel %#x\n", (unsigned)theEvent);

        if (CoalesceEvents && CanCoalesceMouse) {
            if (TraceWindowsEvents)
                printf("Set coalesced mouse value: %d\n", CoalesceMouseValue);
            [theEvent setMouseWheelDelta:CoalesceMouseValue];
        }
        CanCoalesceMouse = NO;

        if (StartWithFirstResponder) {
            responder = [self firstResponder];
        }
        else {
            /* Call on the internal NSFrameView, which is the contentView's superview. */
            responder = [[[self contentView] superview] hitTest:[theEvent locationInWindow]];
        }

        for (; responder != nil; responder = [responder nextResponder])
        {
            if ([responder respondsToSelector:@selector(mouseWheel:)]) {
                if ([responder mouseWheel:theEvent])
                    return YES;
            }
        }

        return NO;
    }
    else if ([theEvent type] == NSApplicationDefined && [theEvent subtype] == VerticalScrollSubtype)
    {
        id responder = [[[self contentView] superview] hitTest:[theEvent locationInWindow]];
        ScrollType scrollType = [theEvent scrollType];

        if (TraceWindowsEvents)
            printf("Handling vertical scroll\n");

        if (CoalesceEvents && CanCoalesce[scrollType] && scrollType != ScrollMax) {
            if (TraceWindowsEvents)
                printf("Set coalesced vertical scroll value type %d to %d\n", scrollType, CoalesceValue[scrollType]);
            [theEvent setScrollIncrement:CoalesceValue[scrollType]];
        }
        CanCoalesce[scrollType] = NO;

        for (; responder != nil; responder = [responder nextResponder])
        {
            if ([responder respondsToSelector:@selector(verticalScroll:)]) {
                if ([responder verticalScroll:theEvent])
                    return YES;
            }
        }
        return NO;
    }
    else if ([theEvent type] == NSApplicationDefined && [theEvent subtype] == HorizontalScrollSubtype)
    {
        id responder = [[[self contentView] superview] hitTest:[theEvent locationInWindow]];
        ScrollType scrollType = [theEvent scrollType];

        if (TraceWindowsEvents)
            printf("Handling horizontal scroll\n");

        if (CoalesceEvents && CanCoalesce[scrollType] && scrollType != ScrollMax) {
            if (TraceWindowsEvents)
                printf("Set coalesced horizontal scroll value type %d to %d\n", scrollType, CoalesceValue[scrollType]);
            [theEvent setScrollIncrement:CoalesceValue[scrollType]];
        }
        CanCoalesce[scrollType] = NO;

        for (; responder != nil; responder = [responder nextResponder])
        {
            if ([responder respondsToSelector:@selector(horizontalScroll:)]) {
                if ([responder horizontalScroll:theEvent])
                    return YES;
            }
        }
        return NO;
    }
    else
    {
        return windowSendEvent(self, _cmd, theEvent);
    }
}

@end


/** Functions to play runtime games.  Slightly munged from Omni's stuff. **/

#import <objc/objc-runtime.h>

static void _RCAddMethod(IMP methodImp, Class class, const char *methodTypes, SEL selector)
{
    struct objc_method_list *newMethodList;

    newMethodList = (struct objc_method_list *)NSZoneMalloc(NSDefaultMallocZone(), sizeof(struct objc_method_list));

    newMethodList->method_count = 1;
    newMethodList->method_list[0].method_name = selector;
    newMethodList->method_list[0].method_imp = methodImp;
    newMethodList->method_list[0].method_types = (char *)methodTypes;
    class_addMethods(class, newMethodList);
}

static IMP RCOverrideInstanceMethodWithSelector(Class aClass, SEL oldSelector, SEL newSelector)
{
    struct objc_method *oldMethod, *newMethod;
    IMP                 oldImp = NULL;

    if ((oldMethod = class_getInstanceMethod(aClass, oldSelector))) {
        if (newMethod = class_getInstanceMethod(aClass, newSelector)) {
            oldImp = oldMethod->method_imp;
            _RCAddMethod(newMethod->method_imp, aClass, oldMethod->method_types, oldSelector);
        }
    }

    return oldImp;
}                                                                 
